/*
    mKEYB.C - minimum keyboard handler for international keyboards

    requires only ~512 byte of preciious memory to do what it does.

    for details, please see readme.txt


    Copyright (C) 2002 by tom.ehlert@ginko.de
*/


#include <stdio.h>
#include <dos.h>

#include "mKeyb.h"




#if defined (LANGUAGE_GR)
    #include "keydefGR.h"
#elif defined (LANGUAGE_SP)
    #include "keydefSP.h"
#elif defined (LANGUAGE_NL)
    #include "keydefNL.h"
#elif defined (LANGUAGE_RU)
    #include "keydefRU.h"
#elif defined (LANGUAGE_UK)
    #include "keydefUK.h"
#elif defined (LANGUAGE_SV)
    #include "keydefSV.h"
#elif defined (LANGUAGE_NO)
    #include "keydefNO.h"
#elif defined (LANGUAGE_SU)
    #include "keydefSU.h"
#elif defined (LANGUAGE_DK)
    #include "keydefDK.h"
#elif defined (LANGUAGE_PO)
    #include "keydefPO.h"
#elif defined (LANGUAGE_IT)
    #include "keydefIT.h"
#elif defined (LANGUAGE_LA)
    #include "keydefLA.h"
#elif defined (LANGUAGE_SF)
    #include "keydefSF.h"
#elif defined (LANGUAGE_SG)
    #include "keydefSG.h"
#else
    #error "you must define a language LANGUAGE_GR,LANGUAGE_RU,..."
#endif



#define DBGprintf printf

void fmemcpy(void far *d, void far *s, uint len)
{
    unsigned char far *dc = d,far *sc = s;
    for (; len; len--)
        *dc++ = *sc++;
}

/*
    this allocates memory as undisturbing as possible.
    strategy choosen:
    allocate at end of memory, in upper memory if present
*/

AllocHighMemory(uint residentsize, int makeResident)
{
    union REGS regs;
    uint allocSeg;
    uint oldlink,oldstrategy;

    regs.x.ax = 0x5802;     // get UMB link state
    int86(0x21,&regs,&regs);
    oldlink = regs.h.al;

    regs.x.ax = 0x5800;     // get allocation strategy
    int86(0x21,&regs,&regs);
    oldstrategy = regs.h.al;


    regs.x.ax = 0x5803;     // set UMB link state
    regs.x.bx = 0x0001;
    int86(0x21,&regs,&regs);

    regs.x.ax = 0x5801;     // set allocation strategy
    regs.x.bx = 0x82;       // last fit, try high, then low
    int86(0x21,&regs,&regs);




    regs.x.ax = 0x4800;                       // dosAllocMem
    regs.x.bx = (residentsize + 15) / 16;       // size in paragraphs
    int86(0x21,&regs,&regs);

    if (regs.x.cflag) { printf("can't allocate memory\n");exit(1); }

    allocSeg = regs.x.ax;


    if (makeResident)
        {
        *(short far *)MK_FP(allocSeg-1,1) = allocSeg;   /* makes it selfreferencing */
        fmemcpy(MK_FP(allocSeg-1,8),"mKEYB   ",8);      /* mark our name */
        }

    regs.x.ax = 0x5803;     // reset UMB link state
    regs.x.bx = oldlink;
    int86(0x21,&regs,&regs);

    regs.x.ax = 0x5801;         // set allocation strategy
    regs.x.bx = oldstrategy;    // last fit, try high, then low
    int86(0x21,&regs,&regs);


    return allocSeg;
}

/*
    VerifyScancodeTableForCorrectness();
    this is in theory not needed.
    in practice, it's better to chack, that the tables are setup correctly
    at least as there are yet many missing keyboard definitions
*/

void VerifyScancodeTableForCorrectness(int print)
{
    extern unsigned char RESIDENT nibble_length[];
    unsigned scancode,flags;
    unsigned char *tbl,*tblnext,*lasttbl;


    for (   tbl = (unsigned char near*)&scancodetable;
                tbl[0] != 0;
                tbl = tblnext)
        {
        if (print)
            printf("offset %02x scancode %02x flags %02x\n",
                tbl-scancodetable, tbl[0], tbl[1]);


        flags = tbl[1];

        tblnext = tbl+nibble_length[flags & ALLSIZEFLAGS];

        if (tblnext > scancodetable+sizeof(scancodetable))
            {
            printf("scancode table error <SIZE> offset 0x%2x, scancode = 0x%02x, flags = 0x%02x\n",
                tbl-scancodetable, tbl[0],flags);
            goto error;
            }

        if (tblnext[0])
            if (tblnext[1] & ~ALLPOSSIBLEFLAGS)
                {
                printf("scancode table error <FLAGS> offset 0x%2x, scancode = 0x%02x, flags = 0x%02x\n",
                    tbl-scancodetable, tbl[0],flags);
                goto error;
                }
        lasttbl = tbl;
        }

    if (tbl + 1 != scancodetable+sizeof(scancodetable))
        {
        printf("scancode table error: early zero found at offset 0x%2x last scan 0x%02x\n",
                tbl-scancodetable, lasttbl[0]);
        goto error;
        }
return;

error:
    if (!print)
        VerifyScancodeTableForCorrectness(1);
    exit(1);
}


main(int argc, char *argv[])
{
    union  REGS regs;
    struct SREGS sregs;
    uint residentSeg;
    uint residentsize;
    uint residentloc;

    uint GOTSR = argc == 1;
    uint endscancodetable;
    int  i;

    if (argv);

    printf("mKEYB 0.20 [" __DATE__ "] - " LANGUAGE " - " STATUS
            " (c) 2002 tom ehlert\n");




    {  // make sure some assumptions old
    int err = 0;

    if (FP_SEG(int2f_handler) != FP_SEG(int15_handler)) err |= 0x0001;
    if (FP_SEG(ResidentScancodetable)  != FP_SEG(int15_handler)) err |= 0x0002;
    if (FP_OFF(ResidentScancodetable)  > 0x800)                  err |= 0x0004;


    if (err)
        {
        printf("compile time error %x\n", err);
        printf("please check your compiler switches and recompile\n");
        return 1;
        }
    }

    VerifyScancodeTableForCorrectness(0);




    /* install the resident part */
    {
    static struct {
        uint   RESIDENT *pos;
        unsigned char *table;
        uint   tablesize;
        } tableoftables[] =
        {
            { NULL,                    scancodetable, sizeof(scancodetable) },

#ifdef COMBI
            { &ResidentCombiTables[0], combi1table,   sizeof(combi1table) },
            { &ResidentCombiTables[1], combi2table,   sizeof(combi2table) },
            { &ResidentCombiTables[2], combi3table,   sizeof(combi3table) },
            { &ResidentCombiTables[3], combi4table,   sizeof(combi4table) },
            { &ResidentCombiTables[4], combi5table,   sizeof(combi5table) }
#endif
        };



        residentsize = FP_OFF(ResidentScancodetable);       /* this MUST be last thing in MAP */

        for (i = 0; i < LENGTH(tableoftables); i++)
            {
            if (tableoftables[i].pos)
                *tableoftables[i].pos = residentsize;       /* remember where we will put it */
            residentsize += tableoftables[i].tablesize;
            }

#ifdef DECIMALDINGSBUMS
        DecimalDingsBums = DECIMALDINGSBUMS;
#endif

#ifdef DEFAULT_LAYOUT_ENGLISH
        usebiosonly_flag = 0;
#endif





        residentSeg = AllocHighMemory(residentsize,GOTSR);  /* fetch enough memory for the TSR part */


#define RESPTR(x) MK_FP(residentSeg, (void near *)x)


        OldInt15 = getvect(0x15);       /* do this before copying */
        OldInt2F = getvect(0x2f);


                                        /* copy code and data up into (high) memory */
        fmemcpy(RESPTR(0),
                MK_FP(FP_SEG(int15_handler),0),
                FP_OFF(ResidentScancodetable));


        residentsize = FP_OFF(ResidentScancodetable);       /* this MUST be last thing in MAP */

        for (i = 0; i < LENGTH(tableoftables); i++)
            {
            fmemcpy(RESPTR(residentsize),tableoftables[i].table,tableoftables[i].tablesize);
            residentsize += tableoftables[i].tablesize;
            }




        regs.x.ax  = 0x2515;                        // dosSetVect
        regs.x.dx  = FP_OFF(int15_handler);
        sregs.ds   = residentSeg;
        int86x(0x21,&regs,&regs,&sregs);

        regs.x.ax  = 0x252f;                        // dosSetVect
        regs.x.dx  = FP_OFF(int2f_handler);
        sregs.ds   = residentSeg;
        int86x(0x21,&regs,&regs,&sregs);

    } /* done with install */


    if (GOTSR)
        {
        exit(0);
        }


    printf("\n mKEYB installed at segment %04x, %u bytes\n",residentSeg, residentsize);

    printf("\n mKEYB debug mode - hit ESC key to terminate test mode\n");

    {
    /* debug only */
    extern uchar far debug_scancode;        /* debug */

    while (*(uchar far *)RESPTR(&debug_scancode) != 1)
        {
        union REGS r;

        if (*(uchar far *)RESPTR(&debug_scancode) != 0)
            {
            printf("scancode    %02x\n",*(uchar far *)RESPTR(&debug_scancode));

            if (*(uchar far *)RESPTR(&debug_scancode) & 0x80 && /* release key ??*/
                *(uchar far *)RESPTR(&debug_scancode) != 0xe0)
                printf("\n");

            *(uchar far *)RESPTR(&debug_scancode) = 0;
            }

        r.x.ax = 0x1100;
        int86(0x16,&r,&r);
        if ((r.x.flags & 0x40) == 0)
            {
            r.x.ax = 0x1000;
            int86(0x16,&r,&r);

            printf("key pressed %04x '%c'\n",r.x.ax,r.h.al);
            }
        }
    }

    setvect(0x15,OldInt15);
    setvect(0x2f,OldInt2F);


    printf("MKEYB - uninstalled\n");

    return 0;

}


/*
    TE - some size optimizations for __TURBOC__

    as printf() is redefined in PRF.C to use no stream functions,
    rather calls DOS directly, these Stream operations are nowhere used,
    but happen to be in the executable.

    so we define some dummy functions here to save some precious bytes :-)

    this is in no way necessary, but saves us some 1500 bytes
*/

#ifdef __TURBOC__

    #define UNREFERENCED_PARAMETER(x) if (x);

    int     _Cdecl flushall (void){return 0;}

    int     _Cdecl fprintf  (FILE *__stream, const char *__format, ...)
                             { UNREFERENCED_PARAMETER (__stream);
                               UNREFERENCED_PARAMETER ( __format);    return 0;}
    int     _Cdecl fseek    (FILE *__stream, long __offset, int __whence)
                             { UNREFERENCED_PARAMETER (__stream);
                               UNREFERENCED_PARAMETER (__offset);
                               UNREFERENCED_PARAMETER ( __whence);
                               return 0;}

    int     _Cdecl setvbuf  (FILE *__stream, char *__buf, int __type, size_t __size)
                             { UNREFERENCED_PARAMETER (__stream);
                               UNREFERENCED_PARAMETER ( __buf);
                               UNREFERENCED_PARAMETER ( __type);
                               UNREFERENCED_PARAMETER ( __size);   return 0;}

    void    _Cdecl _xfflush (void){}
    void    _Cdecl _setupio (void){}

#endif

